/*->c.bf */

#include <stdafx.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "os.h"
#include "flex.h"
#include "alloc.h"

#include "fsx.h"
#include "etc.h"
#include "err.h"


#include "bf.h"




/* BUFFCRIT is a flag set if p has been moved outside the buffer */


/* dump the current buffer full */

os_error * bf_dump(buffer * bf)
{
 os_error * err;

 err=NULL;


/* dprintf(6,"bf_dump in"); */

 if(bf->flags & BUFFMOD)
 {
  if(bf->length)
  {
   {
             err=fs_seek(bf->fh,bf->base);
    if(!err) err=fs_write(bf->fh,bf->data,bf->length);
   }
  }
  bf->flags&=~BUFFMOD;
 }

/* dprintf(6,"bf_dump out"); */

 return(err);
}



/* load the current buffer from current pointer */

static os_error * bf_load(buffer * bf)
{
 os_error * err;

 {
  err=fs_seek(bf->fh,bf->p);
  if(!err)
  {
   bf->base=bf->p;
   err=fs_readn(bf->fh,bf->data,bf->size,&bf->length);
  }
 }

 bf->flags&=~BUFFCRIT;

 return(err);
}


static os_error * bf_init(buffer * bf)
{

/* 
 bf->fh=0;
 bf->p=0;
 bf->flags=0;
 bf->length=0;
 bf->base=0;
*/

 memset(bf,0,sizeof(buffer));

 return(NULL);
}


static os_error * bf_alloc(buffer * bf,int size)
{
 os_error * err;

 err=bf_init(bf);

 do
 {
  err=salloc((flex_ptr)&bf->data,size);
  if(!err)
  {
   bf->size=size;
   break;
  }
  size/=2;
 } while(size>128);

 return(err);
}


static os_error * bf_dealloc(buffer * bf,os_error * ep)
{
 os_error * err;

 err=sfree((flex_ptr)&bf->data);

 if(ep) return(ep);
 else   return(err);
}


os_error * bf_seek(buffer * bf,int posn)
{
 if(posn<bf->base || posn>(bf->base+bf->length)) bf->flags|=BUFFCRIT;

/*
 if(posn>(bf->base+bf->length))
 {
  bf->length=bf->p-bf->base;
  if(bf->length>bf->size) bf->length=bf->size;
 }
*/

 bf->p=posn;
 return(NULL);
}


os_error * bf_seekrel(buffer * bf,int delta)
{
 return(bf_seek(bf,bf->p+delta));
}


os_error * bf_tell(buffer * bf,int * posn)
{
 *posn=bf->p;
 return(NULL);
}



os_error * bf_getc(buffer * bf,int * c)
{
 os_error * err;

 err=NULL;

 if(bf->flags & BUFFCRIT)
 {
  err=bf_dump(bf);
  if(!err) err=bf_load(bf);
 }

 if(bf->p<(bf->base+bf->length))
 {
  *c=bf->data[bf->p-bf->base];
  bf->p++;
  if(bf->p>=(bf->base+bf->length)) bf->flags|=BUFFCRIT;
 }
 else *c=EOF;

 return(err);
}


os_error * bf_putc(buffer * bf,int c)
{
 os_error * err;

 err=NULL;

 if(bf->flags & BUFFCRIT)
 {
  err=bf_dump(bf);    /* NB even if error, update pointers */
  bf->length=0;       /* so that future putc don't crash   */
  bf->base=bf->p;
  bf->flags&=~BUFFCRIT;
 }

 bf->data[bf->p-bf->base]=(char)c;
 bf->flags|=BUFFMOD;
 bf->p++;

 if(bf->p>=(bf->base+bf->size))   bf->flags|=BUFFCRIT;
 if(bf->p>=(bf->base+bf->length)) bf->length++;

 return(err);
}





os_error * bf_readn(buffer * bf,void * b,int len,int * read)
{
 os_error * err;
 char     * p;
 int        left;
 int        chunk;
 int        n=0;    // compiler
 int        eof;

 err=NULL;
 left=len;
 eof=0;
 p=(char*)b;

 if(left)
 {
  if(!(bf->flags & BUFFCRIT))
  {
   /* see how much we can supply from the buffer */

   chunk=bf->length-(bf->p-bf->base);
   if(chunk>0)
   {
    if(chunk>left) chunk=left;

    XMEMCPYX(p,bf->data+(bf->p-bf->base),chunk);
    p+=chunk;

    bf->p+=chunk;
    if(bf->p>=(bf->base+bf->length)) bf->flags|=BUFFCRIT;

    left-=chunk;
   }
  }

  if(left)
  {
   if(left>bf->size) /* try to read integer number of buffers direct full */
   {
    chunk=(left/bf->size);
    chunk*=bf->size;

    {
     err=fs_seek(bf->fh,bf->p);
     if(!err)
     {
      err=fs_readn(bf->fh,p,chunk,&n);
     }
    }

    if(n<chunk) eof=1;

    p+=n;

    bf->p+=n;
    if(bf->p>=(bf->base+bf->length)) bf->flags|=BUFFCRIT;

    left-=n;
   } 

   if(left && !eof && !err)
   {
    /* then try to fill the buffer */

    if(bf->flags & BUFFCRIT) err=bf_dump(bf);
    if(!err) err=bf_load(bf);

    /* then complete the read with what is in the buffer */

    if(!err)
    {
     chunk=bf->length-(bf->p-bf->base);
     if(chunk>0)
     {
      if(chunk>left) chunk=left;

      XMEMCPYX(p,bf->data+(bf->p-bf->base),chunk);
      p+=chunk;

      bf->p+=chunk;
      if(bf->p>=(bf->base+bf->length)) bf->flags|=BUFFCRIT;

      left-=chunk;
     }
    }
   }
  }
 }
 *read=len-left;
 return(err);
}



os_error * bf_read(buffer * bf,void * b,int len)
{
 os_error * err;
 int        n;
 err=bf_readn(bf,b,len,&n);
 if(!err && n!=len) err=geterror(ERD);
 return(err);
}



os_error * bf_writen(buffer * bf,void * b,int len,int * write)
{
 os_error * err;
 char     * p;
 int        left;
 int        chunk;
 int        n=0;    // compiler

 err=NULL;
 left=len;
 p=(char*)b;


/* dprintf(6,"bf_writen in"); */

 if(left)
 {
  if(!(bf->flags & BUFFCRIT))
  {
   /* see how much we can write to the buffer */

   chunk=bf->size-(bf->p-bf->base);
   if(chunk>0)
   {
    if(chunk>left) chunk=left;

    XMEMCPYX(bf->data+(bf->p-bf->base),p,chunk);
    bf->flags|=BUFFMOD;

    p+=chunk;

    bf->p+=chunk;
    if(bf->p>=(bf->base+bf->size))  bf->flags|=BUFFCRIT;
    if(bf->p>(bf->base+bf->length)) bf->length=bf->p-bf->base;

    left-=chunk;
   }
  }

  if(bf->flags & BUFFCRIT)
  {
   err=bf_dump(bf);
   bf->length=0;
   bf->base=bf->p;
   bf->flags&=~BUFFCRIT;
  }

  if(!err && left)
  {
   if(left>bf->size) /* try to write integer number of buffers direct full */
   {
    chunk=(left/bf->size);
    chunk*=bf->size;

    {
     err=fs_seek(bf->fh,bf->p);
     if(!err)
     {
      err=fs_writen(bf->fh,p,chunk,&n);
     }
    }

    p+=n;

    bf->p+=n;
    bf->length=0;
    bf->base=bf->p;
    bf->flags&=~BUFFCRIT;

    left-=n;
   } 

   if(!err && left)
   {
    /* then complete the write to the buffer */

    if(!err)
    {
     chunk=bf->size-(bf->p-bf->base);
     if(chunk>0)
     {
      if(chunk>left) chunk=left;

      XMEMCPYX(bf->data+(bf->p-bf->base),p,chunk);
      bf->flags|=BUFFMOD;

      p+=chunk;

      bf->p+=chunk;
      if(bf->p>=(bf->base+bf->size))  bf->flags|=BUFFCRIT;
      if(bf->p>(bf->base+bf->length)) bf->length=bf->p-bf->base;

      left-=chunk;
     }
    }
   }
  }
 }

 *write=len-left;

/* dprintf(6,"bf_writen out *write=%d",*write); */

 return(err);
}



os_error * bf_write(buffer * bf,void * b,int len)
{
 os_error * err;
 int        n;
 err=bf_writen(bf,b,len,&n);
 if(!err && n!=len) err=geterror(EWR);
 return(err);
}



os_error * bf_geti(buffer * bf,int * c)
{
 return(bf_read(bf,c,sizeof(int)));
}


os_error * bf_puti(buffer * bf,int c)
{
 return(bf_write(bf,&c,sizeof(int)));
}


os_error * bf_gets(buffer * bf,int * c)
{
 *c=0;
 return(bf_read(bf,c,2));
}


os_error * bf_puts(buffer * bf,int c)
{
 return(bf_write(bf,&c,2));
}


os_error * bf_open(char * filename,int mode,int size,buffer * bf)
{
 os_error * err;

 err=bf_alloc(bf,size);
 if(!err)
 {
  bf->flags|=BUFFCRIT;

  err=fs_open(filename,mode,&bf->fh);

  if(err) err=bf_dealloc(bf,err);
 }

 return(err);
}




os_error * bf_close(buffer * bf,os_error * ep)
{
 os_error * err;

 err=bf_dump(bf);
 err=fs_close(bf->fh,err);
 err=bf_dealloc(bf,err);

 if(ep) return(ep);
 else   return(err);
}




os_error * bf_closec(buffer * bf,os_error * ep,char * name,int type)
{
 os_error * err;

 err=bf_close(bf,ep);


 return(err);
}


